import os

from nopecha.nopecha import (
    Recognition,
    Token,
    Balance,
)
from nopecha.error import NopeCHAError, InvalidRequestError, IncompleteJobError, RateLimitError, AuthenticationError, InsufficientCreditError, UnknownError, Timeout, APIError, ServiceUnavailableError

api_base = os.environ.get("NOPECHA_API_BASE", "https://api.nopecha.com")
api_key = os.environ.get("NOPECHA_API_KEY")
proxy = None

__all__ = [
    "Recognition",
    "Token",
    "Balance",

    "NopeCHAError",
    "InvalidRequestError",
    "IncompleteJobError",
    "RateLimitError",
    "AuthenticationError",
    "InsufficientCreditError",
    "UnknownError",
    "Timeout",
    "APIError",
    "ServiceUnavailableError",

    "api_base",
    "api_key",
    "proxy",
]



import json
import platform
import threading
from json import JSONDecodeError
from urllib.parse import urlencode, urlsplit, urlunsplit

import requests

import nopecha
from nopecha import error

TIMEOUT_SECS = 600
MAX_CONNECTION_RETRIES = 2

_thread_context = threading.local()


api_key_to_header = (lambda key: {"Authorization": f"Bearer {key}"})


def _build_api_url(url, query):
    scheme, netloc, path, base_query, fragment = urlsplit(url)

    if base_query:
        query = "%s&%s" % (base_query, query)

    return urlunsplit((scheme, netloc, path, query, fragment))


def _requests_proxies_arg(proxy):
    """Returns a value suitable for the 'proxies' argument to 'requests.request."""
    if proxy is None:
        return None
    elif isinstance(proxy, str):
        return {"http": proxy, "https": proxy}
    elif isinstance(proxy, dict):
        return proxy.copy()
    else:
        raise ValueError("'nopecha.proxy' must be specified as either a string URL or a dict with string URL under the https and/or http keys.")


def _make_session():
    s = requests.Session()
    proxies = _requests_proxies_arg(nopecha.proxy)
    if proxies:
        s.proxies = proxies
    s.mount("https://", requests.adapters.HTTPAdapter(max_retries=MAX_CONNECTION_RETRIES))
    return s


class APIRequestor:
    def __init__(self, key=None, api_base=None):
        self.api_base = api_base or nopecha.api_base
        self.api_key = key or nopecha.api_key

    def request(self, method, url, params=None, headers=None, request_timeout=None):
        result = self.request_raw(method.lower(), url, params=params, supplied_headers=headers, request_timeout=request_timeout)
        resp, rbody, rcode, rheaders = self._interpret_response(result)
        if "error" in resp:
            return self.handle_error_response(rbody, rcode, resp, rheaders)
        return resp

    def handle_error_response(self, rbody, rcode, resp, rheaders):
        error_data = f"{resp['message']}" if "message" in resp else f"{resp['error']}"
        if rcode == 429:
            return error.RateLimitError(error_data, rbody, rcode, resp, rheaders)
        elif rcode == 400:
            return error.InvalidRequestError(error_data, rbody, rcode, resp, rheaders)
        elif rcode == 401:
            return error.AuthenticationError(error_data, rbody, rcode, resp, rheaders)
        elif rcode == 403:
            return error.InsufficientCreditError(error_data, rbody, rcode, resp, rheaders)
        elif rcode == 409:
            return error.IncompleteJobError(error_data, rbody, rcode, resp, rheaders)
        else:
            return error.UnknownError(error_data, rbody, rcode, resp, rheaders)

    def request_headers(self, extra):
        user_agent = "NopeCHA PythonBindings"
        uname_without_node = " ".join(v for k, v in platform.uname()._asdict().items() if k != "node")
        ua = {
            "httplib": "requests",
            "lang": "python",
            "lang_version": platform.python_version(),
            "platform": platform.platform(),
            "publisher": "nopecha",
            "uname": uname_without_node,
        }
        headers = {
            "X-NopeCHA-Client-User-Agent": json.dumps(ua),
            "User-Agent": user_agent,
        }
        headers.update(api_key_to_header(self.api_key))
        headers.update(extra)
        return headers

    def _validate_headers(self, supplied_headers):
        headers = {}
        if supplied_headers is None:
            return headers
        if not isinstance(supplied_headers, dict):
            raise TypeError("Headers must be a dictionary")
        for k, v in supplied_headers.items():
            if not isinstance(k, str):
                raise TypeError("Header keys must be strings")
            if not isinstance(v, str):
                raise TypeError("Header values must be strings")
            headers[k] = v
        return headers

    def request_raw(self, method, url, *, params=None, supplied_headers=None, request_timeout=None):
        abs_url = "%s%s" % (self.api_base, url)
        headers = self._validate_headers(supplied_headers)

        data = None
        if method == "get":
            if params:
                encoded_params = urlencode([(k, v) for k, v in params.items() if v is not None])
                abs_url = _build_api_url(abs_url, encoded_params)
        elif method == "post":
            if params:
                data = json.dumps(params).encode()
                headers["Content-Type"] = "application/json"
        else:
            raise error.APIError(f"Unrecognized HTTP method ${method}. This may indicate a bug in the NopeCHA bindings. Please contact support@nopecha.com for assistance.")

        headers = self.request_headers(headers)

        if not hasattr(_thread_context, "session"):
            _thread_context.session = _make_session()
        try:
            result = _thread_context.session.request(method, abs_url, headers=headers, data=data, timeout=request_timeout if request_timeout else TIMEOUT_SECS)
        except requests.exceptions.Timeout as e:
            raise error.Timeout("Request timed out") from e
        except requests.exceptions.RequestException as e:
            raise error.APIError("Error communicating with NopeCHA") from e
        return result

    def _interpret_response(self, result):
        return self._interpret_response_line(result.content, result.status_code, result.headers)

    def _interpret_response_line(self, rbody, rcode, rheaders):
        if rcode == 503:
            raise error.ServiceUnavailableError("The server is overloaded or not ready yet.", rbody, rcode, headers=rheaders)
        try:
            if hasattr(rbody, "decode"):
                rbody = rbody.decode("utf-8")
            resp = json.loads(rbody)
        except (JSONDecodeError, UnicodeDecodeError):
            raise error.APIError(f"HTTP code {rcode} from API ({rbody})", rbody, rcode, headers=rheaders)
        return resp, rbody, rcode, rheaders



import json
from copy import deepcopy

from nopecha import api_requestor


class NopeCHAObject(dict):
    api_base_override = None

    def __init__(self, api_key=None, api_base=None, **params):
        super(NopeCHAObject, self).__init__()

        self._retrieve_params = params
        object.__setattr__(self, "api_key", api_key)
        object.__setattr__(self, "api_base_override", api_base)

    def __setattr__(self, k, v):
        if k[0] == "_" or k in self.__dict__:
            return super(NopeCHAObject, self).__setattr__(k, v)
        self[k] = v
        return None

    def __getattr__(self, k):
        if k[0] == "_":
            raise AttributeError(k)
        try:
            return self[k]
        except KeyError as err:
            raise AttributeError(*err.args)

    def __delattr__(self, k):
        if k[0] == "_" or k in self.__dict__:
            return super(NopeCHAObject, self).__delattr__(k)
        else:
            del self[k]

    def __setitem__(self, k, v):
        if v == "":
            raise ValueError(f"You cannot set {k} to an empty string. We interpret empty strings as None in requests. You may set {str(self)}.{k} = None to delete the property")
        super(NopeCHAObject, self).__setitem__(k, v)

    def __delitem__(self, k):
        raise NotImplementedError("del is not supported")

    def __setstate__(self, state):
        self.update(state)

    def __reduce__(self):
        return (type(self), (self.api_key,), dict(self))

    @classmethod
    def api_base(cls):
        return None

    def request(self, method, url, params=None, headers=None, request_timeout=None):
        if params is None:
            params = self._retrieve_params
        requestor = api_requestor.APIRequestor(key=self.api_key, api_base=self.api_base_override or self.api_base())
        return requestor.request(method, url, params=params, headers=headers, request_timeout=request_timeout)

    def __repr__(self):
        ident_parts = [type(self).__name__]
        obj = self.get("object")
        if isinstance(obj, str):
            ident_parts.append(obj)
        return "<%s at %s> JSON: %s" % (
            " ".join(ident_parts),
            hex(id(self)),
            str(self),
        )

    def __str__(self):
        obj = self.to_dict_recursive()
        return json.dumps(obj, sort_keys=True, indent=2)

    def to_dict(self):
        return dict(self)

    def to_dict_recursive(self):
        d = dict(self)
        for k, v in d.items():
            if isinstance(v, NopeCHAObject):
                d[k] = v.to_dict_recursive()
            elif isinstance(v, list):
                d[k] = [e.to_dict_recursive() if isinstance(e, NopeCHAObject) else e for e in v]
        return d

    def __copy__(self):
        copied = NopeCHAObject(self.api_key)
        copied._retrieve_params = self._retrieve_params
        for k, v in self.items():
            super(NopeCHAObject, copied).__setitem__(k, v)
        return copied

    def __deepcopy__(self, memo):
        copied = self.__copy__()
        memo[id(self)] = copied
        for k, v in self.items():
            super(NopeCHAObject, copied).__setitem__(k, deepcopy(v, memo))
        return copied



class NopeCHAError(Exception):
    def __init__(self, message=None, http_body=None, http_status=None, json_body=None, headers=None):
        super(NopeCHAError, self).__init__(message)

        self._message = message
        self.http_body = http_body
        self.http_status = http_status
        self.json_body = json_body
        self.headers = headers or {}

    def __str__(self):
        return f"{type(self).__name__}: {self._message}" or "<empty message>"


class InvalidRequestError(NopeCHAError):
    pass


class IncompleteJobError(NopeCHAError):
    pass


class RateLimitError(NopeCHAError):
    pass


class AuthenticationError(NopeCHAError):
    pass


class InsufficientCreditError(NopeCHAError):
    pass


class UnknownError(NopeCHAError):
    pass


class Timeout(NopeCHAError):
    pass


class APIError(NopeCHAError):
    pass


class ServiceUnavailableError(NopeCHAError):
    pass



import time

import nopecha.error
from nopecha.nopecha_object import NopeCHAObject


class NopeCHARequest(NopeCHAObject):
    @classmethod
    def get_url(self):
        raise NotImplementedError()

    @classmethod
    def get_retries(self):
        raise NotImplementedError()

    @classmethod
    def get_interval(self):
        raise NotImplementedError()

    @classmethod
    def post(cls, **params):
        return cls().request("post", cls.get_url(), params)

    @classmethod
    def get(cls, **params):
        for _ in range(cls.get_retries()):
            time.sleep(cls.get_interval())
            result = cls().request("get", cls.get_url(), params)
            if isinstance(result, nopecha.error.IncompleteJobError):
                continue
            return result
        raise nopecha.error.Timeout('Failed to get results')

    @classmethod
    def solve(cls, **params):
        r = cls.post(**params)
        if isinstance(r, nopecha.error.NopeCHAError):
            raise r
        r = cls.get(id=r['data'])
        if isinstance(r, nopecha.error.NopeCHAError):
            raise r
        return r['data']


class Recognition(NopeCHARequest):
    @classmethod
    def get_url(self):
        return "/"

    @classmethod
    def get_retries(self):
        return 120

    @classmethod
    def get_interval(self):
        return 1


class Token(NopeCHARequest):
    @classmethod
    def get_url(self):
        return "/token"

    @classmethod
    def get_retries(self):
        return 180

    @classmethod
    def get_interval(self):
        return 1


class Balance(NopeCHAObject):
    @classmethod
    def get_url(self):
        return "/status"

    @classmethod
    def get(cls, **params):
        r = cls().request("get", cls.get_url(), params)
        if isinstance(r, nopecha.error.NopeCHAError):
            raise r
        return r
